#include <stdio.h>
#include <stdlib.h>

typedef struct __node_t {
    int         key;
    struct __node_t     *next;
} node_t;

typedef struct __list_t {
    node_t      *head;
    lock_t      *lock;    
} list_t;

void List_Init(list_t *L) {
    L->head = NULL;
}

void List_Insert(list_t *L, int key) {
    node_t *new = malloc(sizeof(node_t));
    if (new == NULL) { perror("malloc"); return; }
    new->key  = key;
    // protect this critical section to avoid lost updates
    lock_acquire(L->lock);
    new->next = L->head;
    L->head   = new;
    lock_release(L->lock);
}

void List_Remove(list_t *list, int key) {
    node_t *tmp = L->head;
    note_t *prev = NULL;
    while (tmp) {
        if (tmp->key == key) {
            if (prev) {
                prev->next = tmp->next;
            }
            if (tmp == L->head) {
                L->head = tmp->next;
            }
            free(tmp);
        }
        prev = tmp;
        tmp = tmp->next;
    }
}

int List_Lookup(list_t *L, int key) {
    lock_acquire(L->lock); // needed when List_Remove exists
    node_t *tmp = L->head;
    while (tmp) {
        if (tmp->key == key) {
            // would also need to release here!
            // otherwise lookup could return still holding the lock
            return 1;
        }
        tmp = tmp->next;
    }
    lock_release(L->lock); // needed when List_Remove exists
    return 0;
}

void List_Print(list_t *L) {
    lock_acquire(L->lock); // needed when List_Remove exists
    node_t *tmp = L->head;
    while (tmp) {
        printf("%d ", tmp->key);
        tmp = tmp->next;
    }
    printf("\n");
    lock_release(L->lock); // needed when List_Remove exists
}
